"""
This modules provides a basic Django admin site instance :py:const:`django_tasks.admin.site` that will display the
:py:class:`django_tasks.models.RegisteredTask`, :py:class:`django_tasks.models.DocTask` models, and that is capable of
implementing background task actions on any configured Django model (database table).
"""
import logging

from django.contrib import admin
from django.conf import settings
from django.db.models import QuerySet
from django.http import HttpRequest

from rest_framework.authtoken.models import Token
from rest_framework.authtoken.admin import TokenAdmin

from django_tasks import models
from django_tasks.admin_tools import AdminTaskAction, ChannelTasksAdminSite
from django_tasks.serializers import DocTaskSerializer
from django_tasks.typing import WSResponseJSON


#: The admin site instance to be configured and deployed.
site = ChannelTasksAdminSite()


@admin.register(models.RegisteredTask, site=site)
class RegisteredTaskModelAdmin(admin.ModelAdmin):
    def has_change_permission(self, request, obj=None):
        """
        A registered task must correctly correspond to a task function definition, so this must always
        return `False` to ban introducing inconsistencies.
        """
        return False

    def has_add_permission(self, request):
        """Task registration must be performed on task function definition, so this always returns `False`."""
        return False


site.register(Token, TokenAdmin)


if settings.DEBUG:
    @AdminTaskAction('django_tasks.tasks.doctask_access_test', description='Test async database access')
    def doctask_access_test(modeladmin: admin.ModelAdmin,
                            request: HttpRequest,
                            queryset: QuerySet,
                            ws_response: WSResponseJSON):
        objects_repr = str(queryset) if queryset.count() > 1 else str(queryset.first())
        logging.getLogger('django').info(
            'Requested to run DB access test on %s. Received response: %s.', objects_repr, ws_response)

    @AdminTaskAction('django_tasks.tasks.doctask_deletion_test', description='Test async database DELETE')
    def doctask_deletion_test(modeladmin: admin.ModelAdmin,
                              request: HttpRequest,
                              queryset: QuerySet,
                              ws_response: WSResponseJSON):
        objects_repr = str(queryset) if queryset.count() > 1 else str(queryset.first())
        logging.getLogger('django').info(
            'Requested to delete %s. Received response: %s.', objects_repr, ws_response)


@admin.register(models.DocTask, site=site)
class DocTaskModelAdmin(admin.ModelAdmin):
    list_display = ('registered_task', 'inputs', 'duration', *DocTaskSerializer.Meta.read_only_fields)

    if settings.DEBUG:
        actions = [doctask_access_test, doctask_deletion_test]

    def has_change_permission(self, request, obj=None):
        """Doc tasks must only be modified internally, so this always returns `False`."""
        return False

    def has_add_permission(self, request):
        """Doc tasks must only be created by task scheduling, so this always returns `False`."""
        return False



import asyncio
import collections
import pprint

import bs4
import pytest

from django.test.client import Client

from bdd_coder import decorators
from bdd_coder import tester

from django_tasks.task_runner import TaskRunner


from django_tasks.behaviour.tests.request_cases import AsgiRequestResponseCase, WsgiRequestResponseCase
from django_tasks.behaviour.tests.websocket_test_client import TestingWebSocketClient
from django_tasks.websocket.backend_client import BackendWebSocketClient


response_cases = collections.defaultdict(list)


@pytest.mark.django_db
class BddTester(tester.BddTester):
    """
    The BddTester subclass of this tester package.
    It manages scenario runs. All test classes inherit from this one,
    so generic test methods for this package are expected to be defined here
    """
    gherkin = decorators.Gherkin(logs_path='bdd_runs.log')
    runner = TaskRunner.get()

    task_durations = [0.995, 0.95, 0.94, 0.8]
    credentials = dict(username='Alice', password='AlicePassWd')

    @pytest.fixture(autouse=True)
    def setup_ws_client(self, event_loop):
        timeout = 7
        self.local_ws_client = BackendWebSocketClient(timeout=timeout)
        self.testing_ws_client = TestingWebSocketClient(timeout)
        self.event_collection_task = self.testing_ws_client.collect_events(event_loop)

    @pytest.fixture(autouse=True)
    def setup_django(self, settings):
        self.settings = settings

        from django_tasks import models, wsgi
        self.models = models
        self.wsgi = wsgi

        self.client = Client()

    def assert_admin_call(self, method, path, expected_http_code, data=None):
        bytes_data = '&'.join([f'{k}={v}' for k, v in (data or {}).items()]).encode()
        response = getattr(self.client, method.lower())(
            path=path, data=bytes_data, content_type='application/x-www-form-urlencoded', follow=True,
        )
        assert response.status_code == expected_http_code

        return response

    def assert_rest_api_call(self, method, uri, expected_http_code, data=None):
        self.client.logout()

        case = WsgiRequestResponseCase(method, uri, data)
        case.perform()
        response_cases[f'wsgi-{case.action}'].append(case)

        assert case.response.status_code == expected_http_code, case.response.content.decode()

        return case.response

    async def assert_async_rest_api_call(self, method, uri, expected_http_code, data=None):
        case = AsgiRequestResponseCase(method, uri, data)
        case.perform()
        response_cases[f'asgi-{case.action}'].append(case)

        assert case.response.status_code == expected_http_code, case.response.content.decode()

        return case.response

    async def fake_task_coro_ok(self, duration):
        await asyncio.sleep(duration)
        return duration

    async def fake_task_coro_raise(self, duration):
        await asyncio.sleep(duration)
        raise Exception('Fake error')

    def get_all_admin_messages(self, soup):
        return {k: self.get_admin_messages(soup, k) for k in ('success', 'warning', 'info')}

    @staticmethod
    def get_admin_messages(soup, message_class):
        return [li.contents[0] for li in soup.find_all('li', {'class': message_class})]

    @staticmethod
    def get_soup(content):
        return bs4.BeautifulSoup(content.decode(), features='html.parser')

    async def cancelled_error_success_messages_are_broadcasted(self):
        cancelled, error, success = map(int, self.param)
        self.testing_ws_client.expected_events = {
            'started': cancelled + error + success,
            'cancelled': cancelled, 'error': error, 'success': success,
        }
        timeout = 2
        try:
            await asyncio.wait_for(self.event_collection_task, timeout)
        except TimeoutError:
            self.testing_ws_client.wsapp.close()
            raise AssertionError(
                f'Timeout in event collection. Expected counts: {self.testing_ws_client.expected_events}. '
                f'Collected events in {timeout}s: {pprint.pformat(self.testing_ws_client.events)}.')
        else:
            self.testing_ws_client.expected_events = {}



import logging

from django_tasks.settings import SettingsJson


CHANNEL_TASKS = SettingsJson()

DEBUG = CHANNEL_TASKS.debug
SECRET_KEY = CHANNEL_TASKS.secret_key
ALLOWED_HOSTS = CHANNEL_TASKS.allowed_hosts

LANGUAGE_CODE = CHANNEL_TASKS.language_code
TIME_ZONE = CHANNEL_TASKS.time_zone

DATABASES = CHANNEL_TASKS.databases
DEFAULT_AUTO_FIELD = CHANNEL_TASKS.default_auto_field

INSTALLED_APPS = CHANNEL_TASKS.installed_apps

MIDDLEWARE = CHANNEL_TASKS.middleware

TEMPLATES = CHANNEL_TASKS.templates

CHANNEL_LAYERS = CHANNEL_TASKS.channel_layers
CACHES = CHANNEL_TASKS.caches

AUTHENTICATION_BACKENDS = CHANNEL_TASKS.authentication_backends
AUTH_PASSWORD_VALIDATORS = CHANNEL_TASKS.auth_password_validators

REST_FRAMEWORK = CHANNEL_TASKS.rest_framework

REQUEST_LOGGING_DATA_LOG_LEVEL = logging.INFO
LOGGING = CHANNEL_TASKS.logging

STATIC_URL = CHANNEL_TASKS.static_url
STATIC_ROOT = CHANNEL_TASKS.static_root
MEDIA_URL = CHANNEL_TASKS.media_url
MEDIA_ROOT = CHANNEL_TASKS.media_root

(EMAIL_HOST,
 EMAIL_PORT,
 EMAIL_USE_TLS,
 EMAIL_HOST_USER,
 EMAIL_HOST_PASSWORD) = CHANNEL_TASKS.email_settings



# Generated by Django 4.2.6 on 2023-10-06 03:24

from django.db import migrations, models
import django.db.models.deletion
from django.utils import timezone
import django_tasks.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='RegisteredTask',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('dotted_path', models.CharField(max_length=80, unique=True)),
            ],
        ),
        migrations.CreateModel(
            name='DocTask',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('scheduled_at', models.DateTimeField(default=timezone.now)),
                ('completed_at', models.DateTimeField(null=True)),
                ('inputs', models.JSONField(default=dict, encoder=django_tasks.models.DefensiveJsonEncoder)),
                ('document', models.JSONField(default=list, encoder=django_tasks.models.DefensiveJsonEncoder)),
                ('registered_task', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_tasks.registeredtask')),
            ],
        ),
    ]



import logging

from asgiref.sync import sync_to_async

from rest_framework.authentication import TokenAuthentication
from rest_framework.exceptions import AuthenticationFailed


class AsyncTokenAuthentication(TokenAuthentication):
    @sync_to_async
    def authenticate_credentials(self, key: str):
        return super().authenticate_credentials(key)

    @sync_to_async
    def authenticate(self, request):
        return super().authenticate(request)


class DRFTokenAuthMiddleware:
    """ASGI application middleware that tries to authenticate by DRF Token."""

    def __init__(self, asgi_app):
        self.asgi_app = asgi_app
        self.token_auth = AsyncTokenAuthentication()

    async def __call__(self, scope, receive, send):
        token_key = self.get_token_key(scope)

        if token_key:
            try:
                user, token = await self.token_auth.authenticate_credentials(token_key)
            except AuthenticationFailed:
                logging.getLogger('django').warning('Invalid token %s.', token_key)
            else:
                scope['user'] = user

        asgi_app = await self.asgi_app(scope, receive, send)
        return asgi_app

    def get_token_key(self, scope):
        for name, value in scope.get('headers', []):
            if name == b'authorization':
                auth_header = value.decode().split()
                if len(auth_header) == 2 and auth_header[0] == self.token_auth.keyword:
                    return auth_header[1]



"""
This module provides the base :py:class:`django.contrib.admin.AdminSite` class featuring background task management,
along with the tools for scheduling tasks as Django Admin actions.
"""
import asyncio
import functools
import logging
import os

from asgiref.sync import sync_to_async
from typing import Any, Callable

from django.apps import apps
from django.conf import settings
from django.contrib import admin, messages
from django.db.models import Model, QuerySet
from django.http import HttpRequest

from django_tasks.task_cache import TaskCache
from django_tasks.websocket.backend_client import BackendWebSocketClient
from django_tasks.typing import WSResponseJSON


class ChannelTasksAdminSite(admin.AdminSite):
    """
    The base :py:class:`django.contrib.admin.AdminSite` class, providing the necessary context to connect the
    ASGI unit through websocket in order to handle task event notifications.
    """
    site_title = 'Django Channel Tasks Admin'
    site_header = 'Channel Tasks'
    index_title = 'Index'

    def each_context(self, request: HttpRequest):
        """
        Overrides the Django method by adding to the returned context the web-socket URL of the backgroung task unit
        and the cached task events of the user, if any; note this method may be called from the log-in page.
        """
        context = super().each_context(request)
        username = getattr(request.user, 'username')

        if username and request.user.is_authenticated:
            context['cached_task_events'] = TaskCache(username).get_index()
            context['websocket_uri'] = os.path.join('/', settings.CHANNEL_TASKS.proxy_route, 'tasks/clear-cache')
            context['websocket_port'] = os.getenv('CHANNEL_TASKS_ASGI_PORT', 8001)

        return context


class ModelTask:
    """
    Callable class whose instances are task coroutines that run concurrently on Django querysets, taking a
    list of database IDs as single argument.
    """

    def __init__(self, app_name: str, model_name: str, instance_task: Callable[[Model], Any]):
        """
        Constructs a coroutine task function that will run the given database-sync function concurrently
        on a set of instances of the specified model.

        :param name: Identifier of the Django App that defines the model.
        :param model_name: Name of the model class.
        :param instance_task: Synchronous task function to run on each instance.
        """
        self.model_class = apps.get_model(app_name, model_name)
        self.instance_task = instance_task

    async def __call__(self, instance_ids: list[int]) -> Any:
        """Runs the coroutine concurrently on the corresponding model instances, and returns the list of outputs."""
        logging.getLogger('django').info(
            'Running %s on %s objects %s...',
            self.instance_task.__name__, self.model_class.__name__, instance_ids,
        )
        outputs = await asyncio.gather(*[self.run(pk) for pk in instance_ids])
        return outputs

    async def run(self, instance_id: int) -> Any:
        try:
            instance = await self.model_class.objects.aget(pk=instance_id)
        except self.model_class.DoesNotExist:
            logging.getLogger('django').error(
                'Instance of %s with pk=%s not found.', self.model_class.__name__, instance_id)
        else:
            try:
                output = await sync_to_async(self.instance_task)(instance)
            except Exception:
                logging.getLogger('django').exception('Got exception:')
            else:
                return output


class AdminTaskAction:
    """
    Callable class whose instances are decorators that produce `django.contrib.admin` action functions that
    schedule, through web-socket, background tasks that will run on the Django queryset selected by the admin user.
    """

    def __init__(self, task_name: str, **kwargs):
        """Constructor.

        :param task_name: Dotted path of a registered task that must take a list of database IDs as single argument.
        :param kwargs: Keyword arguments that are passed directly to the `django.contrib.admin.action` decorator.
        """
        self.task_name = task_name
        self.kwargs = kwargs
        self.client = BackendWebSocketClient()

    def __call__(self,
                 post_schedule_callable: Callable[[admin.ModelAdmin, HttpRequest, QuerySet, WSResponseJSON], Any]
                 ) -> Callable[[admin.ModelAdmin, HttpRequest, QuerySet], Any]:
        """Decorator function.

        :param post_schedule_callable: The decorated function. It takes the same arguments as a regular
          `django.contrib.admin` action function plus the JSON object obtained as response to the web-socket
          schedule request.
        """
        @admin.action(**self.kwargs)
        @functools.wraps(post_schedule_callable)
        def action_callable(modeladmin: admin.ModelAdmin, request: HttpRequest, queryset):
            objects_repr = str(queryset) if queryset.count() > 1 else str(queryset.first())
            ws_response = self.client.perform_request('schedule', [dict(
                registered_task=self.task_name,
                inputs={'instance_ids': list(queryset.values_list('pk', flat=True))}
            )], headers={'Cookie': request.headers['Cookie']})
            description = self.kwargs.get('description', self.task_name)
            msg = f"Requested to '{description}' on {objects_repr}."
            modeladmin.message_user(request, msg, messages.INFO)

            return post_schedule_callable(modeladmin, request, queryset, ws_response)

        return action_callable



"""This module defines Django views for deployments with Nginx."""
import os

from django import http, urls, views
from django.contrib.auth.decorators import login_required


class NginxAccelRedirectView(views.View):
    #: These headers are cleared-out so that Nginx can serve their own
    nginx_headers = ['Content-Type',
                     'Content-Disposition',
                     'Accept-Ranges',
                     'Set-Cookie',
                     'Cache-Control',
                     'Expires']
    location = ''

    @classmethod
    def get_url_path(cls, location: str, **kwargs):
        assert location, 'A prefix is required'
        return urls.path(os.path.join(location, '<path:path>').strip('/'),
                         login_required(cls.as_view(location=location)),
                         **kwargs)

    def get(self, request, path, *args, **kwargs):
        response = http.HttpResponse()

        for ignored_header in self.nginx_headers:
            del response[ignored_header]

        response['X-Accel-Redirect'] = f"/_{self.location.strip('/')}/{path}"

        return response


class NginxFileView(NginxAccelRedirectView):
    def get(self, request, path, *args, **kwargs):
        response = super().get(request, path, *args, **kwargs)
        response['Content-Disposition'] = f'attachment; filename="{path}"'
        return response



"""Gathers and names websocket close codes (Websocket ref: https://datatracker.ietf.org/doc/html/rfc6455)."""
import enum

import websocket


class WSCloseCode(enum.IntEnum):
    NORMAL = websocket.STATUS_NORMAL
    GOING_AWAY = websocket.STATUS_GOING_AWAY
    PROTOCOL_ERROR = websocket.STATUS_PROTOCOL_ERROR
    UNSUPPORTED_DATA_TYPE = websocket.STATUS_UNSUPPORTED_DATA_TYPE
    STATUS_NOT_AVAILABLE = websocket.STATUS_STATUS_NOT_AVAILABLE
    ABNORMAL_CLOSURE = websocket.STATUS_ABNORMAL_CLOSED
    INVALID_PAYLOAD = websocket.STATUS_INVALID_PAYLOAD
    POLICY_VIOLATION = websocket.STATUS_POLICY_VIOLATION
    MESSAGE_TOO_BIG = websocket.STATUS_MESSAGE_TOO_BIG
    INVALID_EXTENSION = websocket.STATUS_INVALID_EXTENSION
    UNEXPECTED_CONDITION = websocket.STATUS_UNEXPECTED_CONDITION
    SERVICE_RESTART = 1012
    TRY_AGAIN_LATER = 1013
    BAD_GATEWAY = websocket.STATUS_BAD_GATEWAY
    TLS_HANDSHAKE_ERROR = websocket.STATUS_TLS_HANDSHAKE_ERROR
    UNAUTHORIZED = 3000
    FORBIDDEN = 3003
    TIMEOUT = 3008



"""Module that defines the custom `typing` types of this project, and some type assertion functions."""
from typing import Any, TypeAlias, TypedDict


#: Union type of all the flat JSON-serializable objects, lacking substructure.
FlatJSON: TypeAlias = str | float | int | bool | None

#: Generic, recursive JSON-serializable type.
JSON: TypeAlias = dict[str, 'JSON'] | list['JSON'] | FlatJSON

#: Generic JSON-serializable type for consumer events.
EventJSON = TypedDict('EventJSON', {'type': str, 'content': JSON})


#: JSON-serializable type for the task status data broadcasted by the task runner.
TaskStatusJSON = TypedDict('TaskStatusJSON', {
    'registered_task': str,
    'status': str,
    'http_status': int,
    'exception-repr': str,
    'output': JSON,
}, total=False)

#: JSON-serializable type for the content of task events broadcasted by the task runner.
TaskMessageJSON = TypedDict('TaskMessageJSON', {'task_id': str, 'detail': TaskStatusJSON})

#: Type for the web-socket responses.
WSResponseJSON = TypedDict('WSResponseJSON', {
    'request_id': str,
    'http_status': int,
    'details': list[JSON],
}, total=False)

#: JSON-serializable type for cache-clear request content.
CacheClearJSON = TypedDict('CacheClearJSON', {'task_id': str})

#: JSON-serializable type for a single-task schedule request content.
TaskJSON = TypedDict('TaskJSON', {'registered_task': str, 'inputs': dict[str, JSON]})

#: JSON-serializable type for a DocTask schedule request content.
DocTaskJSON = TypedDict('DocTaskJSON', {'id': int, 'registered_task': str, 'inputs': dict[str, JSON]})


def is_string_key_dict(value: Any) -> bool:
    return isinstance(value, dict) and all(isinstance(k, str) for k in value)


def is_string_key_dict_list(value: Any) -> bool:
    return isinstance(value, list) and all(is_string_key_dict(v) for v in value)


def is_string_list(value: Any) -> bool:
    return isinstance(value, list) and all(isinstance(v, str) for v in value)



import abc
import json

import requests

from rest_framework.test import APIClient


from django_tasks.http_status import HttpStatus
from django_tasks.typing import JSON


def get_test_credential(name: str):
    with open(f'.test_{name}.txt') as secret_file:
        return secret_file.read().strip()


class RequestResponseCase(metaclass=abc.ABCMeta):
    """An HTTP request-response example, for testing and documentation purposes."""

    #: The base URL for all requests.
    base_url: str = ''

    #: Headers included in all requests.
    default_request_headers: dict[str, str] = {
        'Authorization': f'Token {get_test_credential("token")}',
        'Content-Type': 'application/json',
    }

    def __init__(self, method: str, uri: str, data: JSON = None, **headers: str):
        self.method = method.strip().lower()
        self.uri = uri.strip()
        self.data = data
        self.headers = headers
        self.headers.update(self.default_request_headers)
        self.perform()

    @abc.abstractmethod
    def perform_request(self):
        """Perform the request, return the response."""

    @property
    def response(self):
        return self._response

    @property
    def status_code(self) -> HttpStatus:
        return self._status_code

    @property
    def response_json(self) -> JSON:
        return self._response_json

    def perform(self):
        """Performs the request and sets all required properties from the obtained response."""
        self._response = self.perform_request()
        self._status_code = HttpStatus(self.response.status_code)
        self._response_json = (json.loads(self.response.content)
                               if self.response.headers.get('content-type') == 'application/json'
                               else None)

    @property
    def url(self) -> str:
        return f'{self.base_url}/{self.uri}'

    @property
    def action(self) -> str:
        return f"{self.method}-{(self.uri.split('?')[0] if '?' in self.uri else self.uri).replace('/', '')}"

    def generate_rst_lines(self, name: str | int = 'example'):
        yield f'*Request {name}*:\n\n'
        yield '.. sourcecode:: http\n\n'
        yield f'   {self.method.upper()} /{self.uri} HTTP/1.1\n'

        for key, header in self.headers.items():
            yield f'   {key.capitalize()}: {header}\n'

        if self.data:
            yield '   \n'
            for data_line in json.dumps(self.data, indent=4).splitlines():
                yield f'   {data_line}\n'

        yield f'\n*Response {name}*:\n\n'
        yield '.. sourcecode:: http\n\n'
        yield f'   HTTP/1.1 {self.status_code} {self.status_code.name}\n'

        for key, header in self.response.headers.items():
            yield f'   {key.capitalize()}: {header}\n'

        if self.response_json:
            yield '   \n'
            for data_line in json.dumps(self.response_json, indent=4).splitlines():
                yield f'   {data_line}\n'
        elif self.response.content:
            yield '   \n'
            yield f'   {self.response.content}\n'


class AsgiRequestResponseCase(RequestResponseCase):
    base_url = 'http://127.0.0.1:8001'

    def perform_request(self):
        return getattr(requests, self.method)(self.url, data=json.dumps(self.data).encode(), headers=self.headers)


class WsgiRequestResponseCase(RequestResponseCase):
    base_url = ''

    def perform_request(self):
        return getattr(APIClient(), self.method)(self.url, data=self.data, headers=self.headers)


class HttpEndpointCaseSet:
    def __init__(self, file_name: str, *request_response_cases: RequestResponseCase):
        self.file_name = file_name
        self.cases = request_response_cases

    @property
    def rst_path(self) -> str:
        return f"sphinx/source/requests/{self.file_name}.rst"

    def write_rst(self):
        with open(self.rst_path, 'w') as rst_file:
            for request_response_case in self.cases:
                rst_file.writelines(request_response_case.generate_rst_lines())
                rst_file.write('\n')



"""
This module defines the consumer classes for background task management.
"""
import abc
import json
import logging
import uuid

from asgiref.sync import sync_to_async

from channels.exceptions import StopConsumer
from channels.consumer import AsyncConsumer
from channels.generic.http import AsyncHttpConsumer
from channels.generic.websocket import AsyncJsonWebsocketConsumer
from django.conf import settings
from rest_framework import status
from rest_framework.exceptions import ValidationError
from rest_framework.settings import api_settings

from django_tasks.serializers import DocTaskSerializer
from django_tasks.scheduler import DocTaskScheduler, schedule_tasks
from django_tasks.task_cache import TaskCache
from django_tasks.websocket.close_codes import WSCloseCode

from django_tasks.typing import JSON, EventJSON, DocTaskJSON, TaskJSON, WSResponseJSON


class TaskGroupConsumer(AsyncConsumer):
    @property
    def user_group(self) -> str:
        """The name of the group of consumers that is assigned to the user."""
        return f"{self.scope['user'].username}_{settings.CHANNEL_TASKS.channel_group}"

    @property
    def request_id(self) -> str:
        """
        The request ID provided in the corresponding header, or a newly generated ID if the header is missing.
        This value will be returned to clients as an identifier of the request; websocket clients must provide
        it in the Request-ID header in order to be able to check the returned ID value in subsequent messages.
        """
        for name, value in self.scope.get('headers', []):
            if name == b'request-id':
                id_value: str = value.decode()
                return id_value

        return uuid.uuid4().hex


class TaskGroupJsonConsumer(TaskGroupConsumer, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def receive_json(self, request_content: JSON) -> int:
        """Process the received, already parsed, JSON request content, and return an HTTP status code."""

    @abc.abstractmethod
    async def send_bad_request_response(self, error: ValidationError) -> None:
        """Send a 400 message in case of validation error."""


class HttpConsumer(AsyncHttpConsumer):
    async def handle(self, body: bytes):
        handler = getattr(self, 'handle_' + self.scope['method'].lower(), None)

        if handler is None:
            await self.send_response(status.HTTP_405_METHOD_NOT_ALLOWED, body=b'')
        else:
            await handler(body)


class TaskScheduleConsumer(TaskGroupJsonConsumer):
    async def receive_json(self, request_content: JSON) -> int:
        """Processes task schedule websocket requests."""
        logging.getLogger('django').debug(
            'Processing task schedule through channel %s. Data: %s', self.channel_name, request_content)
        try:
            many_serializer = await DocTaskSerializer.get_valid_task_group_serializer(request_content)
        except ValidationError as error:
            await self.send_bad_request_response(error)
            return status.HTTP_400_BAD_REQUEST
        else:
            data: list[TaskJSON] = await many_serializer.adata
            await schedule_tasks(self.request_id, self.scope['user'].username, *data)
            return status.HTTP_200_OK


class DocTaskScheduleConsumer(TaskGroupJsonConsumer):
    async def receive_json(self, request_content: JSON) -> int:
        """Processes doc-task schedule websocket requests."""
        logging.getLogger('django').debug(
            'Processing DocTask schedule through channel %s. Data: %s', self.channel_name, request_content)
        try:
            many_serializer, doctasks = await DocTaskSerializer.create_doctask_group(request_content)
        except ValidationError as error:
            await self.send_bad_request_response(error)
            return status.HTTP_400_BAD_REQUEST
        else:
            data: list[DocTaskJSON] = await many_serializer.adata
            await DocTaskScheduler.schedule_doctasks(self.request_id, self.scope['user'].username, *data)
            return status.HTTP_201_CREATED


class TaskCacheClearConsumer(TaskGroupJsonConsumer):
    async def receive_json(self, request_content: JSON) -> int:
        """Clears a specific task cache."""
        logging.getLogger('django').debug(
            'Processing cache clear through channel %s. Data: %s', self.channel_name, request_content)
        await sync_to_async(self.user_task_cache.clear_task_cache)(request_content['task_id'])
        return status.HTTP_200_OK


class TaskWebSocketConsumer(TaskGroupJsonConsumer, AsyncJsonWebsocketConsumer):
    async def send_bad_request_response(self, error: ValidationError) -> None:
        """Broadcasts an HTTP 400 message through the user's group of consumers."""
        content: WSResponseJSON = {
            'http_status': status.HTTP_400_BAD_REQUEST,
            'request_id': self.request_id,
            'details': error.get_full_details(),
        }
        await self.group_send({'type': 'task.badrequest', 'content': content})

    async def group_send(self, event: EventJSON) -> None:
        """Distributes the given `event` through the group of the user of this instance."""
        await self.channel_layer.group_send(self.user_group, event)

    async def task_started(self, event: EventJSON) -> None:
        """Echoes the task.started document."""
        await self.send_json(content=event)

    async def task_success(self, event: EventJSON) -> None:
        """Echoes the task.success document."""
        await self.send_json(content=event)

    async def task_cancelled(self, event: EventJSON) -> None:
        """Echoes the task.cancelled document."""
        await self.send_json(content=event)

    async def task_error(self, event: EventJSON) -> None:
        """Echoes the task.error document."""
        await self.send_json(content=event)

    async def task_badrequest(self, event: EventJSON) -> None:
        """Echoes the task.badrequest document."""
        await self.send_json(content=event)

    async def stop_unauthorized(self) -> None:
        """Stops the consumer if the user is not authenticated."""
        if not self.scope['user'].is_authenticated:
            logging.getLogger('django').warning('Unauthenticated user %s. Closing websocket.', self.scope['user'])
            await self.close(code=WSCloseCode.UNAUTHORIZED)
            raise StopConsumer()

    async def connect(self) -> None:
        """Performs authentication and set-up actions on connection."""
        await self.stop_unauthorized()
        await super().connect()
        self.user_task_cache = TaskCache(self.scope['user'].username)
        await self.channel_layer.group_add(self.user_group, self.channel_name)
        logging.getLogger('django').debug(
            'Connected user "%s" through channel %s.', self.scope['user'].username, self.channel_name)

    async def disconnect(self, close_code: int) -> None:
        """Performs clean-up actions on disconnection."""
        await self.channel_layer.group_discard(self.user_group, self.channel_name)
        logging.getLogger('django').debug(
            'Disconnected channel %s. User: %s. Reason: %s',
            self.channel_name, self.scope['user'].username, repr(WSCloseCode(close_code)))


class TaskScheduleWebSocketConsumer(TaskScheduleConsumer, TaskWebSocketConsumer):
    """The websocket consumer for task schedule requests."""


class DocTaskScheduleWebSocketConsumer(DocTaskScheduleConsumer, TaskWebSocketConsumer):
    """The websocket consumer for doc-task schedule requests."""


class CacheClearWebSocketConsumer(TaskCacheClearConsumer, TaskWebSocketConsumer):
    """The websocket consumer for cache-clear requests."""


class TaskJsonHttpConsumer(TaskGroupJsonConsumer, HttpConsumer):
    async def handle_post(self, body: bytes):
        try:
            request_content: JSON = json.loads(body)
        except json.JSONDecodeError as error:
            await self.send_bad_request_response(ValidationError({'non_field_errors': repr(error)}))
        else:
            http_status = await self.receive_json(request_content)
            if http_status < 400:
                content: WSResponseJSON = {'request_id': self.request_id}
                await self.send_json_response(http_status, content)

    async def send_bad_request_response(self, error: ValidationError) -> None:
        """Broadcasts an HTTP 400 message through the user's group of consumers."""
        content: WSResponseJSON = {
            'request_id': self.request_id, 'details': error.get_full_details(),
        }
        await self.send_json_response(status.HTTP_400_BAD_REQUEST, content)

    async def send_json_response(self, status_code: int, content: WSResponseJSON) -> None:
        await self.send_response(status_code, json.dumps(content).encode(),
                                 headers={b'Content-Type': b'application/json'})


class TaskScheduleHttpConsumer(TaskScheduleConsumer, TaskJsonHttpConsumer):
    """The HTTP consumer for task schedule requests."""


class DocTaskScheduleHttpConsumer(DocTaskScheduleConsumer, TaskJsonHttpConsumer):
    """The HTTP consumer for doc-task schedule requests."""



import asyncio
import time

from django.core.management import call_command

from rest_framework import status

from django_tasks.behaviour.tests.request_cases import HttpEndpointCaseSet, get_test_credential
from django_tasks.typing import JSON

from . import base


def teardown_module():
    """
    Called by Pytest at teardown of the test module, employed here to
    log final scenario results, and to write request-response documentation sources.
    """
    base.BddTester.gherkin.log()

    for file_name, cases in base.response_cases.items():
        HttpEndpointCaseSet(file_name, *cases).write_rst()


class TestWebsocketScheduling(base.BddTester):
    """
    This covers:
    * The task runner
    * The tasks websocket API
    """

    @base.BddTester.gherkin()
    def test_several_tasks_are_scheduled_with_ws_message(self):
        """
        When a failed and some OK tasks are scheduled through WS
        Then $(0) cancelled $(1) error $(4) success messages are broadcasted
        """

    def a_failed_and_some_ok_tasks_are_scheduled_through_ws(self):
        name = 'django_tasks.tasks.sleep_test'
        task_data: list[JSON] = [
            dict(registered_task=name, inputs={'duration': dn}) for dn in self.task_durations]
        task_data.append(dict(registered_task=name, inputs={'duration': 0.15, 'raise_error': True}))
        response = self.local_ws_client.perform_request(
            'schedule', task_data, headers={'Cookie': get_test_credential('cookie')})
        assert response['http_status'] == status.HTTP_200_OK


class TestTaskRunner(base.BddTester):
    """
    Several tasks may be scheduled to run concurrently, and their states are broadcasted.
    Task information may also be stored in database.
    This covers:
    * The task runner
    * The websocket broadcasting
    """

    @base.BddTester.gherkin()
    def test_concurrent_error_and_cancellation(self):
        """
        When a `failed`, a `cancelled` and some `OK` tasks are scheduled
        Then completion times do not accumulate
        And $(1) cancelled $(1) error $(4) success messages are broadcasted
        And the different task statuses are correctly stored
        """

    async def a_failed_a_cancelled_and_some_ok_tasks_are_scheduled(self):
        failed_task, cancelled_task, *ok_tasks = await asyncio.gather(
            self.runner.schedule(self.fake_task_coro_raise(0.1)),
            self.runner.schedule(self.fake_task_coro_ok(10)),
            *[self.runner.schedule(self.fake_task_coro_ok(d)) for d in self.task_durations])

        return failed_task, cancelled_task, ok_tasks

    async def completion_times_do_not_accumulate(self):
        initial_time = time.time()
        task_results = await asyncio.gather(*self.get_output('ok'))
        self.get_output('cancelled').cancel()
        elapsed_time = time.time() - initial_time

        assert task_results == self.task_durations
        assert elapsed_time < 1

    async def the_different_task_statuses_are_correctly_stored(self):
        failed_task_info = self.runner.get_task_status(self.get_output('failed'))
        assert failed_task_info['status'] == 'Error'
        assert failed_task_info['exception-repr'].strip() == "Exception('Fake error')"

        await asyncio.sleep(0.01)
        cancelled_task_info = self.runner.get_task_status(self.get_output('cancelled'))
        assert cancelled_task_info['status'] == 'Cancelled'


class TaskAdminUserCreation(base.BddTester):
    """
    As a site administrator,
    I want to create staff users with task management and scheduling permissions,
    so that they can start operating with a temporary password.
    """

    @base.BddTester.gherkin()
    def a_task_admin_is_created_by_command(self):
        """
        When a task admin `user` is created by command
        Then the user has the correct status
        """

    def a_task_admin_user_is_created_by_command(self, django_user_model):
        self.credentials['password'] = call_command(
            'create_task_admin', self.credentials['username'], 'fake@gmail.com'
        )
        user = django_user_model.objects.get(username=self.credentials['username'])

        return user,

    def the_user_has_the_correct_status(self):
        user = self.get_output('user')
        assert user.check_password(self.credentials['password'])
        assert user.is_superuser is False
        assert user.is_staff is True
        assert user.is_active is True

    def the_user_logs_in(self):
        logged_in = self.client.login(**self.credentials)
        assert logged_in


class TestRestApiWithTokenAuth(TaskAdminUserCreation):
    """
    Staff users may obtain a token through Django admin site, and use it to schedule
    concurrent tasks through REST API.
    This covers:
    * The task runner
    * Admin site usage to create API tokens
    * User creation with management command
    * The tasks REST API
    """

    @base.BddTester.gherkin()
    def test_many_tasks_execution_post_with_result_storage(self):
        """
        When a failed and some OK `tasks` are posted
        Then $(0) cancelled $(1) error $(4) success messages are broadcasted
        And the different task results are correctly stored in DB
        """

    @base.BddTester.gherkin()
    def test_single_task_execution_post_with_result_storage(self):
        """
        When a failed `task` is posted with duration $(0.1)
        Then $(0) cancelled $(1) error $(0) success messages are broadcasted
        And the task result is correctly stored in DB
        """

    @base.BddTester.gherkin()
    def test_a_task_admin_creates_an_api_token(self):
        """
        Given a task admin is created by command
        When the user logs in
        Then the user may obtain an API `token`
        """

    @base.BddTester.gherkin()
    def test_many_tasks_post_client_error(self):
        """
        When an authenticated user specifies a task array with several `errors`
        Then a 400 response is returned with all error details
        """

    async def a_failed_and_some_ok_tasks_are_posted(self):
        name = 'django_tasks.tasks.sleep_test'
        task_data = [dict(registered_task=name, inputs={'duration': dn}) for dn in self.task_durations]
        task_data.append(dict(registered_task=name, inputs={'duration': 0.15, 'raise_error': True}))
        response = await self.assert_async_rest_api_call(
            'POST', 'api/doctasks/schedule', status.HTTP_201_CREATED, data=task_data)

        return response.json(),

    async def the_different_task_results_are_correctly_stored_in_db(self):
        response = await self.assert_async_rest_api_call('GET', 'adrf/doctasks?limit=2&offset=1', status.HTTP_200_OK)
        tasks = response.json()
        assert tasks['count'] >= 5

    def a_failed_task_is_posted_with_duration(self):
        duration = float(self.param)
        data = dict(registered_task='django_tasks.tasks.sleep_test',
                    inputs={'duration': duration, 'raise_error': True})
        response = self.assert_rest_api_call('POST', 'api/doctasks', status.HTTP_201_CREATED, data=data)

        return response.json(),

    def the_user_may_obtain_an_api_token(self):
        response = self.assert_admin_call(
            'POST', '/admin/authtoken/token/add/', status.HTTP_200_OK,
            {'user': self.get_output('user').pk, '_save': 'Save'},
        )
        soup = self.get_soup(response.content)

        errors = soup.find('ul', {'class': 'errorlist'})
        assert not errors

        messages = self.get_all_admin_messages(soup)
        assert len(messages['success']) == 1, messages

        return messages['success'][0].split()[2].strip('“”'),

    async def the_task_result_is_correctly_stored_in_db(self):
        response = await self.assert_async_rest_api_call('GET', 'adrf/doctasks?limit=1', status.HTTP_200_OK)
        tasks = response.json()
        assert tasks['count'] >= 1

    async def an_authenticated_user_specifies_a_task_array_with_several_errors(self):
        response = await self.assert_async_rest_api_call(
            'POST', 'api/doctasks/schedule', status.HTTP_400_BAD_REQUEST, data=[
                {
                    "registered_task": "django_tasks.foo.sleep_test",
                    "inputs": {
                        "duration": 4,
                        "raise_error": False,
                    },
                },
                {
                    "registered_task": "django_tasks.tasks.sleep_test",
                    "inputs": {
                        "wrong_key": True,
                    },
                }])

        return response.json(),

    def a_400_response_is_returned_with_all_error_details(self):
        error_details: list[JSON] = self.get_output('errors').get('details', [])
        assert error_details == [
            {
                "registered_task": [
                    {
                        "message": "Object with dotted_path=django_tasks.foo.sleep_test does not exist.",
                        "code": "does_not_exist"
                    }
                ]
            },
            {
                "inputs": [
                    {
                        "message": "Missing required parameters {'duration'}.",
                        "code": "invalid"
                    },
                    {
                        "message": "Unknown parameters {'wrong_key'}.",
                        "code": "invalid"
                    }
                ]
            }
        ]


class TestAsyncAdminSiteActions(TaskAdminUserCreation):
    """
    This covers:
    * The admin tools module
    """

    @base.BddTester.gherkin()
    def test_database_access_async_actions_run_ok(self):
        """
        Given a task admin is created by command
        And the user logs in
        When the user runs the $(doctask_access_test) action
        And the user runs the $(doctask_deletion_test) action
        """

    def the_user_runs_the_action(self):
        self.assert_admin_call('POST', '/admin/django_tasks/doctask/', status.HTTP_200_OK, {
            'action': self.param,
            '_selected_action': self.models.DocTask.objects.first().pk})



from django.apps import apps
from django.core.wsgi import get_wsgi_application

from django_tasks import tasks

application = get_wsgi_application()

RegisteredTask = apps.get_model('django_tasks', 'RegisteredTask')
RegisteredTask.register(tasks.sleep_test)
RegisteredTask.register(tasks.doctask_deletion_test)
RegisteredTask.register(tasks.doctask_access_test)



import inspect
import json
import logging

from typing import Callable

from django.db.models import Model, CharField, DateTimeField, JSONField, ForeignKey, CASCADE
from django.utils import timezone


class DefensiveJsonEncoder(json.JSONEncoder):
    def default(self, obj):
        try:
            return super().default(obj)
        except TypeError:
            return repr(obj)


class RegisteredTask(Model):
    """Specifies a trusted coroutine function to run as task."""
    dotted_path: CharField = CharField(max_length=80, unique=True)

    def __str__(self):
        return self.dotted_path

    @classmethod
    def register(cls, callable: Callable) -> Callable:
        """
        Registers the given coroutine function in database.

        Usage is as follows::

            RegisteredTask = apps.get_model('django_tasks', 'RegisteredTask')
            RegisteredTask.register(new_coroutine_function)

        after Django model setup.

        Raises `AssertionError` if `callable` is not a coroutine function.
        """
        assert inspect.iscoroutinefunction(callable), 'The function must be a coroutine'

        instance, created = cls.objects.get_or_create(
            dotted_path=f'{inspect.getmodule(callable).__spec__.name}.{callable.__name__}'
        )
        msg = 'Registered new task %s' if created else 'Task %s already registered'
        logging.getLogger('django').info(msg, instance)

        return callable


class DocTask(Model):
    """Stored information of a task execution."""
    registered_task: ForeignKey = ForeignKey(RegisteredTask, on_delete=CASCADE)
    scheduled_at: DateTimeField = DateTimeField(default=timezone.now)
    completed_at: DateTimeField = DateTimeField(null=True)
    inputs: JSONField = JSONField(default=dict, encoder=DefensiveJsonEncoder)
    document: JSONField = JSONField(default=list, encoder=DefensiveJsonEncoder)

    def __str__(self):
        return f'Doc-task {self.pk}, ' + (
            f'completed at {self.completed_at}, took {self.duration}' if self.completed_at
            else f'running for {self.duration}')

    @property
    def duration(self):
        return (self.completed_at if self.completed_at else timezone.now()) - self.scheduled_at

    async def on_completion(self, task_info):
        self.completed_at = timezone.now()
        self.document.append(task_info)
        await self.asave()



"""Provides indexed HTTP status codes."""
import enum

from rest_framework import status


class HttpStatus(enum.IntEnum):
    CONTINUE = status.HTTP_100_CONTINUE
    SWITCHING_PROTOCOLS = status.HTTP_101_SWITCHING_PROTOCOLS
    PROCESSING = status.HTTP_102_PROCESSING
    EARLY_HINTS = status.HTTP_103_EARLY_HINTS
    OK = status.HTTP_200_OK
    CREATED = status.HTTP_201_CREATED
    ACCEPTED = status.HTTP_202_ACCEPTED
    NON_AUTHORITATIVE_INFORMATION = status.HTTP_203_NON_AUTHORITATIVE_INFORMATION
    NO_CONTENT = status.HTTP_204_NO_CONTENT
    RESET_CONTENT = status.HTTP_205_RESET_CONTENT
    PARTIAL_CONTENT = status.HTTP_206_PARTIAL_CONTENT
    MULTI_STATUS = status.HTTP_207_MULTI_STATUS
    ALREADY_REPORTED = status.HTTP_208_ALREADY_REPORTED
    IM_USED = status.HTTP_226_IM_USED
    MULTIPLE_CHOICES = status.HTTP_300_MULTIPLE_CHOICES
    MOVED_PERMANENTLY = status.HTTP_301_MOVED_PERMANENTLY
    FOUND = status.HTTP_302_FOUND
    SEE_OTHER = status.HTTP_303_SEE_OTHER
    NOT_MODIFIED = status.HTTP_304_NOT_MODIFIED
    USE_PROXY = status.HTTP_305_USE_PROXY
    RESERVED = status.HTTP_306_RESERVED
    TEMPORARY_REDIRECT = status.HTTP_307_TEMPORARY_REDIRECT
    PERMANENT_REDIRECT = status.HTTP_308_PERMANENT_REDIRECT
    BAD_REQUEST = status.HTTP_400_BAD_REQUEST
    UNAUTHORIZED = status.HTTP_401_UNAUTHORIZED
    PAYMENT_REQUIRED = status.HTTP_402_PAYMENT_REQUIRED
    FORBIDDEN = status.HTTP_403_FORBIDDEN
    NOT_FOUND = status.HTTP_404_NOT_FOUND
    METHOD_NOT_ALLOWED = status.HTTP_405_METHOD_NOT_ALLOWED
    NOT_ACCEPTABLE = status.HTTP_406_NOT_ACCEPTABLE
    PROXY_AUTHENTICATION_REQUIRED = status.HTTP_407_PROXY_AUTHENTICATION_REQUIRED
    REQUEST_TIMEOUT = status.HTTP_408_REQUEST_TIMEOUT
    CONFLICT = status.HTTP_409_CONFLICT
    GONE = status.HTTP_410_GONE
    LENGTH_REQUIRED = status.HTTP_411_LENGTH_REQUIRED
    PRECONDITION_FAILED = status.HTTP_412_PRECONDITION_FAILED
    REQUEST_ENTITY_TOO_LARGE = status.HTTP_413_REQUEST_ENTITY_TOO_LARGE
    REQUEST_URI_TOO_LONG = status.HTTP_414_REQUEST_URI_TOO_LONG
    UNSUPPORTED_MEDIA_TYPE = status.HTTP_415_UNSUPPORTED_MEDIA_TYPE
    REQUESTED_RANGE_NOT_SATISFIABLE = status.HTTP_416_REQUESTED_RANGE_NOT_SATISFIABLE
    EXPECTATION_FAILED = status.HTTP_417_EXPECTATION_FAILED
    IM_A_TEAPOT = status.HTTP_418_IM_A_TEAPOT
    MISDIRECTED_REQUEST = status.HTTP_421_MISDIRECTED_REQUEST
    UNPROCESSABLE_ENTITY = status.HTTP_422_UNPROCESSABLE_ENTITY
    LOCKED = status.HTTP_423_LOCKED
    FAILED_DEPENDENCY = status.HTTP_424_FAILED_DEPENDENCY
    TOO_EARLY = status.HTTP_425_TOO_EARLY
    UPGRADE_REQUIRED = status.HTTP_426_UPGRADE_REQUIRED
    PRECONDITION_REQUIRED = status.HTTP_428_PRECONDITION_REQUIRED
    TOO_MANY_REQUESTS = status.HTTP_429_TOO_MANY_REQUESTS
    REQUEST_HEADER_FIELDS_TOO_LARGE = status.HTTP_431_REQUEST_HEADER_FIELDS_TOO_LARGE
    UNAVAILABLE_FOR_LEGAL_REASONS = status.HTTP_451_UNAVAILABLE_FOR_LEGAL_REASONS
    INTERNAL_SERVER_ERROR = status.HTTP_500_INTERNAL_SERVER_ERROR
    NOT_IMPLEMENTED = status.HTTP_501_NOT_IMPLEMENTED
    BAD_GATEWAY = status.HTTP_502_BAD_GATEWAY
    SERVICE_UNAVAILABLE = status.HTTP_503_SERVICE_UNAVAILABLE
    GATEWAY_TIMEOUT = status.HTTP_504_GATEWAY_TIMEOUT
    HTTP_VERSION_NOT_SUPPORTED = status.HTTP_505_HTTP_VERSION_NOT_SUPPORTED
    VARIANT_ALSO_NEGOTIATES = status.HTTP_506_VARIANT_ALSO_NEGOTIATES
    INSUFFICIENT_STORAGE = status.HTTP_507_INSUFFICIENT_STORAGE
    LOOP_DETECTED = status.HTTP_508_LOOP_DETECTED
    BANDWIDTH_LIMIT_EXCEEDED = status.HTTP_509_BANDWIDTH_LIMIT_EXCEEDED
    NOT_EXTENDED = status.HTTP_510_NOT_EXTENDED
    NETWORK_AUTHENTICATION_REQUIRED = status.HTTP_511_NETWORK_AUTHENTICATION_REQUIRED



"""
This module provides the :py:class:`django_tasks.websocket.backend_client.BackendWebSocketClient` class,
which manages communications between WSGI and ASGI systems.
"""
from __future__ import annotations
from typing import Any, Callable, Optional

import functools
import json
import logging
import uuid

import websocket

from rest_framework import status
from django.conf import settings

from django_tasks.typing import JSON, WSResponseJSON
from django_tasks.websocket.close_codes import WSCloseCode


def catch_websocket_errors(client_method: Callable) -> Callable:
    """
    Method decorator for :py:class:`django_tasks.websocket.backend_client.BackendWebSocketClient`.
    Decorated methods will catch any :py:class:`websocket.WebSocketException` and return it instead of raising.
    """
    @functools.wraps(client_method)
    def safe_client_method(client: BackendWebSocketClient, *args, **kwargs) -> tuple[bool, Any]:
        try:
            return_value = client_method(client, *args, **kwargs)
            return True, return_value
        except websocket.WebSocketException as error:
            return False, error

    return safe_client_method


class BackendWebSocketClient:
    """
    Wrapper for handy usage of :py:class:`websocket.WebSocket` within the backend, able to handle WSGI requests
    asyncronously through websocket, returning the first websocket messages received from the ASGI endpoint for a
    specific request.
    """
    #: URL path for local access to the ASGI endpoint
    local_route: str = (
        'tasks' if not settings.CHANNEL_TASKS.proxy_route else f'{settings.CHANNEL_TASKS.proxy_route}-local/tasks')

    #: The configured URL for local access to the ASGI endpoint.
    local_url: str = f'ws://127.0.0.1:{settings.CHANNEL_TASKS.local_port}/{local_route}/'

    #: Default headers for the ASGI endpoint.
    default_headers: dict[str, str] = {'Content-Type': 'application/json'}

    #: Default websocket timeout, in seconds. It should be as less as possible for quick responses, since
    #: the client will wait this time for new messages.
    default_timeout: float = 0.1

    #: Maximum number of response messages to collect for a request.
    max_response_msg_collect: int = 2

    def __init__(self, **connect_kwargs):
        self.connect_kwargs = connect_kwargs
        self.connect_kwargs.setdefault('timeout', self.default_timeout)
        self.ws = websocket.WebSocket()
        websocket.setdefaulttimeout(self.connect_kwargs['timeout'])

    def __repr__(self):
        return f'<BackendWebSocketClient: {self.connect_kwargs}>'

    def perform_request(self,
                        action: str,
                        content: JSON,
                        headers: Optional[dict[str, str]] = None) -> WSResponseJSON:
        header = headers or {}
        header.update(self.default_headers)
        header['Request-ID'] = uuid.uuid4().hex

        connect_ok, connect_error = self.connect(action, header)
        if not connect_ok:
            return self.bad_gateway_message(header['Request-ID'], connect_error)

        send_ok, send_error = self.send_json(content)
        if not send_ok:
            return self.bad_gateway_message(header['Request-ID'], send_error)

        response = self.get_first_response(header['Request-ID'])
        self.disconnect(
            WSCloseCode.BAD_GATEWAY if response['http_status'] == status.HTTP_502_BAD_GATEWAY
            else WSCloseCode.NORMAL)

        return response

    @staticmethod
    def bad_gateway_message(request_id: str, error: websocket.WebSocketException) -> WSResponseJSON:
        """Constructs and returns the bad gateway message.

        :param request_id: The universal ID of the request that raised the error.
        :param error: The web-socket error raised during the request.
        """
        return {
            'request_id': request_id,
            'http_status': status.HTTP_502_BAD_GATEWAY,
            'details': [repr(error)],
        }

    @catch_websocket_errors
    def connect(self, action: str, header: dict[str, str]):
        """Tries to connect to the corresponding endpoint with the given headers."""
        return self.ws.connect(self.local_url + action, header=header, **self.connect_kwargs)

    @catch_websocket_errors
    def disconnect(self, close_code: int):
        """Tries to disconnect with the given close code."""
        return self.ws.close(status=close_code)

    @catch_websocket_errors
    def send_json(self, json_data: JSON):
        """Tries to send the given JSON data."""
        return self.ws.send(json.dumps(json_data))

    @catch_websocket_errors
    def receive(self):
        """Tries to receive a message."""
        return self.ws.recv()

    def get_first_response(self, request_id: str) -> WSResponseJSON:
        """
        Having performed a request with the given ID, collects and returns the first response messages
        to that request.
        """
        response: WSResponseJSON = {
            'request_id': request_id, 'details': [], 'http_status': status.HTTP_502_BAD_GATEWAY
        }
        http_statuses: list[int] = []

        for _ in range(self.max_response_msg_collect):
            ok, raw_msg = self.receive()

            if not ok or not raw_msg:
                break

            is_response: bool = True
            try:
                msg = json.loads(raw_msg)
            except json.JSONDecodeError:
                is_response = False
            else:
                if msg.get('content', {}).get('request_id') == request_id:
                    bad_request_response: WSResponseJSON = msg['content']
                    return bad_request_response
                elif msg.get('content', {}).get('task_id'):
                    task_request_id, _ = msg['content']['task_id'].split('.')
                    if task_request_id == request_id:
                        logging.getLogger('django').debug(
                            'Received response message to request %s: %s', request_id, msg)
                        http_statuses.append(msg['content']['detail'].pop('http_status'))
                        response['details'].append(msg['content']['detail'])
                    else:
                        is_response = False
                else:
                    is_response = False

            if is_response is False:
                logging.getLogger('django').debug(
                    'Discarded unrelated message, received after request %s: %s', request_id, raw_msg)

        if http_statuses:
            response['http_status'] = max(http_statuses)

        return response



"""This module provides the :py:class:`django_tasks.task_cache.TaskCache` class."""
from collections import defaultdict

import logging

from django.core.cache import cache

from django_tasks.typing import TaskMessageJSON


class TaskCache:
    """Handles the task event cache of a user, using the configured Django cache."""

    def __init__(self, user_name: str):
        self.user_name = user_name

    @property
    def cache_key(self) -> str:
        return f'{self.user_name}.task_events'

    def get_index(self):
        return cache.get_or_set(self.cache_key, defaultdict(list))

    def clear_task_cache(self, task_id: str):
        """Clears a specific task cache, or logs a warning if not found."""
        current_index = self.get_index()

        if task_id in current_index:
            del current_index[task_id]
            cache.set(self.cache_key, current_index)
        else:
            logging.getLogger('django').warning('No cache found for %s.', task_id)

    def cache_task_event(self, task_id: str, event_content: TaskMessageJSON):
        """Stores the given task event data in the user's cache."""
        current_index = self.get_index()
        current_index[task_id].append(event_content)
        cache.set(self.cache_key, current_index)



"""Module providing all URL patterns, after performing the Django setup."""
from django.conf import settings

from rest_framework.routers import SimpleRouter

from django_tasks.django_setup import django
from django_tasks import admin
from django_tasks.viewsets import WSDocTaskViewSet, DocTaskViewSet
from django_tasks.consumers import (
    TaskScheduleHttpConsumer,
    DocTaskScheduleHttpConsumer,
    TaskScheduleWebSocketConsumer,
    DocTaskScheduleWebSocketConsumer,
    CacheClearWebSocketConsumer,
)

path = django.urls.path
re_path = django.urls.re_path


class OptionalSlashRouter(SimpleRouter):
    def __init__(self):
        super().__init__()
        self.trailing_slash = '/?'


def get_wsgi_urls():
    if settings.CHANNEL_TASKS.expose_rest_api is True:
        drf_router = OptionalSlashRouter()
        drf_router.register('doctasks', WSDocTaskViewSet, basename='doctask')
        yield path('api/', django.urls.include(drf_router.urls))

    yield path('admin/', admin.site.urls)


def get_asgi_urls():
    if settings.CHANNEL_TASKS.expose_rest_api is True:
        drf_router = OptionalSlashRouter()
        drf_router.register('adrf/doctasks', DocTaskViewSet, basename='doctask')
        yield from drf_router.urls


def get_http_channels_urls():
    if settings.CHANNEL_TASKS.expose_rest_api is True:
        yield path('doctasks/schedule', DocTaskScheduleHttpConsumer.as_asgi())
        yield path('tasks/schedule', TaskScheduleHttpConsumer.as_asgi())


def get_websocket_urls():
    yield path('tasks/schedule', TaskScheduleWebSocketConsumer.as_asgi())
    yield path('tasks/schedule-store', DocTaskScheduleWebSocketConsumer.as_asgi())
    yield path('tasks/clear-cache', CacheClearWebSocketConsumer.as_asgi())



"""This module provides the DRF view sets, which are employed in ASGI and WSGI endpoints."""
from adrf.viewsets import ModelViewSet as AsyncModelViewSet
from rest_framework.viewsets import ModelViewSet
from rest_framework.decorators import action
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.status import HTTP_200_OK, HTTP_201_CREATED

from django_tasks.models import DocTask
from django_tasks.serializers import DocTaskSerializer
from django_tasks.websocket.backend_client import BackendWebSocketClient


class WSDocTaskViewSet(ModelViewSet):
    """
    DRF model viewset for :py:class:`django_tasks.models.DocTask` that connects to the background task ASGI unit
    through web-socket.
    """
    http_method_names = ['post', 'delete', 'head', 'options', 'trace']
    queryset = DocTask.objects.all()
    serializer_class = DocTaskSerializer

    #: Name of the header to include to authorize against the ASGI application.
    auth_header = 'Authorization'

    #: The :py:class:`django_tasks.websocket.backend_client.BackendWebSocketClient` instance employed by this class.
    ws_client = BackendWebSocketClient()

    def create(self, request: Request, *args, **kwargs):
        """DRF action that schedules a doc-task through local Websocket."""
        ws_response = self.ws_client.perform_request('schedule-store', [request.data], headers={
            self.auth_header: request.headers[self.auth_header],
        })
        status = ws_response.pop('http_status')
        return Response(status=HTTP_201_CREATED if status == HTTP_200_OK else status, data=ws_response)

    @action(detail=False, methods=['post'])
    def schedule(self, request: Request, *args, **kwargs):
        """DRF action that schedules an array of doc-tasks through local Websocket."""
        ws_response = self.ws_client.perform_request('schedule-store', request.data, headers={
            self.auth_header: request.headers[self.auth_header],
        })
        status = ws_response.pop('http_status')
        return Response(status=HTTP_201_CREATED if status == HTTP_200_OK else status, data=ws_response)


class DocTaskViewSet(AsyncModelViewSet):
    """Asynchronous DRF model viewset for :py:class:`django_tasks.models.DocTask`.

    This is currently implemented for fetching operations only.
    """
    http_method_names = ['get', 'head', 'options', 'trace']
    queryset = DocTask.objects.all()
    serializer_class = DocTaskSerializer



from asgiref.sync import sync_to_async

import asyncio
import collections
import json
import logging
import websocket

from django.conf import settings


class TestingWebSocketClient:
    """Wrapper for handy usage of `websocket.WebSocketApp` for testing."""
    local_route = ('tasks' if not settings.CHANNEL_TASKS.proxy_route
                   else f'{settings.CHANNEL_TASKS.proxy_route}-local/tasks')
    local_url = f'ws://127.0.0.1:{settings.CHANNEL_TASKS.local_port}/{local_route}/clear-cache'
    header = {'Content-Type': 'application/json'}

    def __init__(self, timeout=8):
        websocket.setdefaulttimeout(timeout)
        self.wsapp = websocket.WebSocketApp(
            self.local_url, header=self.header, on_message=self.on_message, on_error=self.on_error,
        )
        self.events = collections.defaultdict(list)
        self.expected_events = {}

    def on_message(self, wsapp, message: str):
        logging.getLogger('django').debug('Received local WS message: %s', message)
        task_event = json.loads(message)
        self.events[task_event['status'].lower()].append(task_event)

        if self.expected_events and self.expected_events_collected:
            wsapp.close()

    def on_error(self, wsapp, error: websocket.WebSocketTimeoutException):
        logging.getLogger('django').error('Catched local WS error: %s. Closing connection.', error)
        wsapp.close()

    def collect_events(self, event_loop):
        return asyncio.wrap_future(asyncio.run_coroutine_threadsafe(
            sync_to_async(self.wsapp.run_forever)(), event_loop))

    @property
    def expected_events_collected(self) -> bool:
        # FIXME: should not ignore 'started' status messages
        return all(len(self.events[event_type]) == count
                   for event_type, count in self.expected_events.items() if event_type != 'started')



"""
This module provides the tools for scheduling arrays of tasks, with or without task result storage.
"""
import asyncio
import logging

from asgiref.sync import sync_to_async
from typing import Optional

from django_tasks import models
from django_tasks.task_runner import TaskRunner
from django_tasks.task_inspector import get_task_coro
from django_tasks.typing import TaskStatusJSON, TaskJSON, DocTaskJSON


class DocTaskScheduler:
    """
    Static class that manages all doc-task runs, capable of scheduling, as requested by a Django user,
    arrays of tasks that are documented in database as :py:class:`django_tasks.models.DocTask` instances.
    """
    model = models.DocTask

    #: Will hold all running doc-task database IDs by task ID
    doctask_index: dict[str, int] = {}

    @classmethod
    def retrieve_doctask(cls, task_id: str) -> Optional[models.DocTask]:
        if task_id in cls.doctask_index:
            doctask_id = cls.doctask_index[task_id]
            try:
                return cls.model.objects.get(pk=doctask_id)
            except cls.model.DoesNotExist:
                count = cls.model.objects.count()
                logging.getLogger('django').error(
                    'Memorized doctask ID %s not found in DB, among %s entries.', doctask_id, count)
        return None

    @classmethod
    async def store_doctask_result(cls, task_id: str, task_info: TaskStatusJSON) -> None:
        doctask = await sync_to_async(cls.retrieve_doctask)(task_id)

        if doctask:
            await doctask.on_completion(task_info)
            del cls.doctask_index[task_id]
            logging.getLogger('django').info('Stored %s.', repr(doctask))

    @classmethod
    async def schedule_doctask(cls, task_id: str, user_name: str, valid_data: DocTaskJSON) -> asyncio.Future:
        """Schedules a single task, and stores results in DB."""
        task_coro = get_task_coro(valid_data['registered_task'], valid_data['inputs'])
        runner = TaskRunner.get()
        task = await runner.schedule(
            task_coro.coroutine, cls.store_doctask_result, task_id=task_id, user_name=user_name
        )
        cls.doctask_index[task_id] = valid_data['id']
        logging.getLogger('django').info('Scheduled doc-task %s callable=%s.', valid_data, task_coro.callable)
        return task

    @classmethod
    async def schedule_doctasks(cls, request_id: str, user_name: str, *valid_data: DocTaskJSON) -> list[asyncio.Future]:
        futures = await asyncio.gather(*[
            cls.schedule_doctask(f'{request_id}.{n}', user_name, data) for n, data in enumerate(valid_data)
        ])
        return futures


async def schedule_tasks(request_id: str, user_name: str, *valid_data: TaskJSON) -> list[asyncio.Future]:
    """Schedules an array of tasks as requested by a Django user."""
    runner = TaskRunner.get()
    futures = await asyncio.gather(*[runner.schedule(
        get_task_coro(dat['registered_task'], dat['inputs']).coroutine,
        task_id=f'{request_id}.{n}', user_name=user_name,
    ) for n, dat in enumerate(valid_data)])
    return futures



from django.core.asgi import get_asgi_application

from channels.auth import AuthMiddlewareStack
from channels.routing import ProtocolTypeRouter, URLRouter
from channels.security.websocket import AllowedHostsOriginValidator

from django_tasks.authentication import DRFTokenAuthMiddleware
from django_tasks import urls


adrf_application = get_asgi_application()

application = ProtocolTypeRouter({
    'http': URLRouter([
        urls.re_path(r'^adrf/', adrf_application),
        urls.re_path(r'^api/', DRFTokenAuthMiddleware(
            AuthMiddlewareStack(URLRouter(list(urls.get_http_channels_urls()))))),
    ]),
    'websocket': AllowedHostsOriginValidator(
        DRFTokenAuthMiddleware(AuthMiddlewareStack(URLRouter(list(urls.get_websocket_urls()))))
    ),
})



"""
This module defines the :py:class:`django_tasks.settings.SettingsJson` class, and the Django settings modules
`django_tasks.settings.asgi`, `django_tasks.settings.wsgi` for the corresponding types of deployment; these
modules are intended for both testing and production deployments, and are configured from
a :py:class:`~django_tasks.settings.SettingsJson` instance.
"""
import importlib.util
import json
import os

from django.core.exceptions import ImproperlyConfigured

from django_tasks.typing import JSON, is_string_key_dict, is_string_key_dict_list, is_string_list


class SettingsJson:
    """
    Class in charge of providing Django setting values, as specified in a JSON settings file whose values override
    the default values provided here.
    """
    #: Name of the environment variable that holds the path to the settings JSON file.
    json_key: str = 'CHANNEL_TASKS_SETTINGS_PATH'

    #: Name of the environment variable that holds the Django secret key.
    secret_key_key: str = 'DJANGO_SECRET_KEY'

    #: Name of the channel-tasks Django app.
    channel_tasks_appname: str = 'django_tasks'

    #: List of identifiers of the Django apps that must be intalled. The "install-apps" setting may be
    #: specified with an array of additional apps to install.
    required_installed_apps: list[str] = [
        'django.contrib.auth',
        'django.contrib.contenttypes',
        'django.contrib.sessions',
        'django.contrib.staticfiles',
        'rest_framework.authtoken',
        'adrf',
        'django.contrib.messages',
        'django_filters',
        channel_tasks_appname,
        'django.contrib.admin',
        'django_sass_compiler',
    ]

    #: Default DRF settings, may be overriden with care using the "rest-framework" setting.
    default_drf: dict[str, JSON] = dict(
        DEFAULT_RENDERER_CLASSES=[
            'rest_framework.renderers.JSONRenderer',
        ],
        DEFAULT_PARSER_CLASSES=[
            'rest_framework.parsers.JSONParser',
        ],
        DEFAULT_PAGINATION_CLASS='rest_framework.pagination.LimitOffsetPagination',
        DEFAULT_FILTER_BACKENDS=[
            'django_filters.rest_framework.DjangoFilterBackend',
        ],
        DEFAULT_AUTHENTICATION_CLASSES=[
            'rest_framework.authentication.TokenAuthentication',
        ],
        DEFAULT_PERMISSION_CLASSES=[
            'rest_framework.permissions.IsAuthenticated',
        ],
        TEST_REQUEST_DEFAULT_FORMAT='json',
    )

    #: Sequence of middleware specifications required for secure deployments. The "append-middleware" setting
    #: may be specified with an array of additional middleware.
    required_middleware: list[str] = [
        'django.middleware.security.SecurityMiddleware',
        'django.contrib.sessions.middleware.SessionMiddleware',
        'django.middleware.common.CommonMiddleware',
        'django.middleware.csrf.CsrfViewMiddleware',
        'django.contrib.auth.middleware.AuthenticationMiddleware',
        'django.contrib.messages.middleware.MessageMiddleware',
        'django.middleware.clickjacking.XFrameOptionsMiddleware',
    ]

    #: Default auth password validators, may be overriden with care using "auth-password-validators".
    default_auth_password_validators: list[dict[str, JSON]] = [
        {
            'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
        },
        {
            'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
        },
        {
            'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
        },
        {
            'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
        },
    ]

    #: Default auth backends, may be overriden with care using "authentication-backends".
    default_authentication_backends: list[str] = [
        'django.contrib.auth.backends.ModelBackend',
    ]

    @classmethod
    def get_default_templates(cls) -> list[dict[str, JSON]]:
        """Returns the Django TEMPLATES setting value that will be set by default.

        This value may be overriden, with care, using the "templates" setting."""
        return [
            {
                'BACKEND': 'django.template.backends.django.DjangoTemplates',
                'DIRS': [f"{os.path.dirname(importlib.util.find_spec(cls.channel_tasks_appname).origin)}/templates"],
                'APP_DIRS': True,
                'OPTIONS': {
                    'context_processors': [
                        'django.template.context_processors.debug',
                        'django.template.context_processors.request',
                        'django.contrib.auth.context_processors.auth',
                        'django.contrib.messages.context_processors.messages',
                    ],
                },
            },
        ]

    @staticmethod
    def get_default_logging(log_level: str) -> dict[str, JSON]:
        return dict(
            version=1,
            disable_existing_loggers=False,
            handlers={
                'console': {
                    'class': 'logging.StreamHandler',
                    'formatter': 'thread-logname',
                },
                'console-debug': {
                    'class': 'logging.StreamHandler',
                    'formatter': 'verbose',
                },
            },
            formatters={
                'verbose': {
                    'format': '{levelname} {asctime} {threadName} ({pathname}) {funcName}:L{lineno} ★ {message}',
                    'style': '{',
                },
                'thread-logname': {
                    'format': '{levelname} {asctime} ({threadName}) {name} ★ {message}',
                    'style': '{',
                },
            },
            loggers={
                'django': {
                    'level': log_level,
                    'handlers': ['console-debug'],
                },
                'django.request': {
                    'handlers': ['console'],
                    'level': log_level,
                    'propagate': False,
                },
                'django.channels': {
                    'handlers': ['console'],
                    'level': log_level,
                    'propagate': False,
                },
            },
        )

    def __init__(self):
        self.json_path: str = os.getenv(self.json_key, '')

        if not os.path.isfile(self.json_path):
            raise ImproperlyConfigured(f'Channel-tasks settings file at {self.json_key}={self.json_path} not found.')

        with open(self.json_path) as json_file:
            self.jsonlike: dict[str, JSON] = json.load(json_file)

        if self.secret_key_key not in os.environ:
            raise ImproperlyConfigured(f'Expected a Django secret key in {self.secret_key_key} envvar.')

        self.secret_key: str = os.environ[self.secret_key_key]

    def wrong_type_error(self, key: str, type_repr: str) -> ImproperlyConfigured:
        """
        Constructs and returns a :py:class:`django.core.exceptions.ImproperlyConfigured` exception,
        indicating, with `type_repr`, the required type for the `key` entry.
        """
        return ImproperlyConfigured(f"Setting value for '{key}' must be of type '{type_repr}' in {self.json_path}")

    def get_boolean(self, key: str, default: bool) -> bool:
        """
        Returns a type-checked boolean from the `key` entry,
        or raises :py:class:`django.core.exceptions.ImproperlyConfigured`.
        """
        value = self.jsonlike.get(key, default)

        if not isinstance(value, bool):
            raise self.wrong_type_error(key, 'bool')

        return value

    def get_int(self, key: str, default: int) -> int:
        """
        Returns a type-checked integer from the `key` entry,
        or raises :py:class:`django.core.exceptions.ImproperlyConfigured`.
        """
        value = self.jsonlike.get(key, default)

        if not isinstance(value, int):
            raise self.wrong_type_error(key, 'int')

        return value

    def get_string(self, key: str, default: str) -> str:
        """
        Returns a type-checked string from the `key` entry,
        or raises :py:class:`django.core.exceptions.ImproperlyConfigured`.
        """
        value = self.jsonlike.get(key, default)

        if not isinstance(value, str):
            raise self.wrong_type_error(key, 'str')

        return value

    def get_string_list(self, key: str, default: list[str]) -> list[str]:
        """
        Returns a type-checked list of strings from the `key` entry,
        or raises :py:class:`django.core.exceptions.ImproperlyConfigured`.
        """
        value = self.jsonlike.get(key, default)

        if not is_string_list(value):
            raise self.wrong_type_error(key, 'list[str]')

        return value

    def get_dict(self, key: str, default: dict[str, JSON]) -> dict[str, JSON]:
        """
        Returns a type-checked string-key dictionary from the `key` entry,
        or raises :py:class:`django.core.exceptions.ImproperlyConfigured`.
        """
        value = self.jsonlike.get(key, default)

        if not is_string_key_dict(value):
            raise self.wrong_type_error(key, 'dict[str]')

        return value

    def get_dict_list(self, key: str, default: list[dict[str, JSON]]) -> list[dict[str, JSON]]:
        """
        Returns a type-checked list of string-key dictionaries from the `key` entry,
        or raises :py:class:`django.core.exceptions.ImproperlyConfigured`.
        """
        value = self.jsonlike.get(key, default)

        if not is_string_key_dict_list(value):
            raise self.wrong_type_error(key, 'list[dict[str]]')

        return value

    @property
    def allowed_hosts(self) -> list[str]:
        """Will be set as the Django ALLOWED_HOSTS setting value.

        Only the loopback address, for local websocket connections, and the configured "server-name" will
        be allowed."""
        return ['127.0.0.1', self.server_name]

    @property
    def server_name(self) -> str:
        """The configured "server-name" value. Defaults to localhost."""
        return self.get_string('server-name', 'localhost')

    @property
    def install_apps(self) -> list[str]:
        """The configured "install-apps" value. Defaults to empty array (no additional apps)."""
        return self.get_string_list('install-apps', [])

    @property
    def installed_apps(self) -> list[str]:
        """Will be set as the Django INSTALLED_APPS setting value."""
        return self.required_installed_apps + self.install_apps

    @property
    def debug(self) -> bool:
        """Will be set as the Django DEBUG setting value. Defaults to `False` (the secure value)."""
        return self.get_boolean('debug', False)

    @property
    def proxy_route(self) -> str:
        return str(self.jsonlike.get('proxy-route', ''))

    @property
    def local_port(self) -> int:
        return self.get_int('local-port', 8001)

    @property
    def log_level(self) -> str:
        """The configured "log-level" value. Defaults to INFO."""
        return self.get_string('log-level', 'INFO')

    @property
    def logging(self) -> dict[str, JSON]:
        """Will be set as the Django LOGGING setting value, taking the configured "log-level" setting."""
        return self.get_dict('logging', self.get_default_logging(self.log_level))

    @property
    def middleware(self) -> list[str]:
        """Will be set as the Django MIDDLEWARE setting value, taking the "append-middleware" setting."""
        return self.required_middleware + self.get_string_list('append-middleware', [])

    @property
    def templates(self) -> list[dict[str, JSON]]:
        """Will be set as the Django TEMPLATES setting value."""
        return self.get_dict_list('templates', self.get_default_templates())

    @property
    def language_code(self) -> str:
        """Will be set as the Django LANGUAGE_CODE setting value, from the "language-code" setting.

        Defaults to en-GB."""
        return self.get_string('language-code', 'en-gb')

    @property
    def time_zone(self) -> str:
        """Will be set as the Django TIME_ZONE setting value, from the "time-zone" setting.

        Defaults to UCT."""
        return self.get_string('time-zone', 'UTC')

    @property
    def auth_password_validators(self) -> list[dict[str, JSON]]:
        """Will be set as the Django AUTH_PASSWORD_VALIDATORS setting value."""
        return self.get_dict_list('auth-password-validators', self.default_auth_password_validators)

    @property
    def authentication_backends(self) -> list[str]:
        """Will be set as the Django AUTHENTICATION_BACKENDS setting value."""
        return self.get_string_list('authentication-backends', self.default_authentication_backends)

    @property
    def rest_framework(self) -> dict[str, JSON]:
        """Will be set as the REST_FRAMEWORK setting value."""
        return self.get_dict('rest-framework', self.default_drf)

    @property
    def databases(self) -> dict[str, dict[str, JSON]]:
        default_db = self.get_dict('database', {
            'ENGINE': 'django.db.backends.sqlite3',
            'NAME': 'channel-tasks.sqlite3',
        })
        db_settings: dict[str, dict[str, JSON]] = {'default': {k.upper(): v for k, v in default_db.items()}}
        db_settings['default'].setdefault('PASSWORD', os.getenv('CHANNEL_TASKS_DB_PASSWORD', ''))

        return db_settings

    @property
    def default_auto_field(self) -> str:
        """Will be set as the Django DEFAULT_AUTO_FIELD setting value."""
        return self.get_string('default-auto-field', 'django.db.models.BigAutoField')

    @property
    def channel_layers(self) -> dict[str, JSON]:
        """Will be set as the CHANNEL_LAYERS setting value."""
        return {
            'default': {
                'BACKEND': 'channels_redis.core.RedisChannelLayer',
                'CONFIG': {
                    'hosts': [[self.redis_host, self.redis_port]],
                },
            },
        }

    @property
    def caches(self) -> dict[str, JSON]:
        """Will be set as the CACHES setting value."""
        return {
            'default': {
                'BACKEND': 'django.core.cache.backends.redis.RedisCache',
                'LOCATION': f'redis://{self.redis_host}:{self.redis_port}',
                'TIMEOUT': 4*86400,
            },
        }

    @property
    def redis_host(self) -> str:
        return self.get_string('redis-host', '127.0.0.1')

    @property
    def redis_port(self) -> int:
        return self.get_int('redis-port', 6379)

    @property
    def static_root(self) -> str:
        """Will be set as the Django STATIC_ROOT setting value."""
        return self.get_string('static-root', '/www/django_tasks/static')

    @property
    def media_root(self) -> str:
        """Will be set as the Django MEDIA_ROOT setting value."""
        return self.get_string('media-root', '/www/django_tasks/media')

    @property
    def static_url(self) -> str:
        """Will be set as the Django STATIC_URL setting value."""
        return self.get_string('static-url', '/static/')

    @property
    def media_url(self) -> str:
        """Will be set as the Django MEDIA_URL setting value."""
        return self.get_string('media-url', '/media/')

    @property
    def email_settings(self) -> tuple[str, int, bool, str, str]:
        return (self.email_host,
                self.email_port,
                self.email_use_tls,
                os.getenv('CHANNEL_TASKS_EMAIL_USER', ''),
                os.getenv('CHANNEL_TASKS_EMAIL_PASSWORD', ''))

    @property
    def email_host(self) -> str:
        """Will be set as the Django EMAIL_HOST setting value.

        Defaults to empty string."""
        return self.get_string('email-host', '')

    @property
    def email_port(self) -> int:
        """Will be set as the Django EMAIL_PORT setting value.

        Defaults to 0."""
        return self.get_int('email-port', 0)

    @property
    def email_use_tls(self) -> bool:
        """Will be set as the Django EMAIL_USE_TLS setting value.

        Defaults to `False`."""
        return self.get_boolean('email-use-tls', False)

    @property
    def channel_group(self) -> str:
        """Channel-tasks setting: name for the background task consumer groups.

        Defaults to 'tasks'."""
        return self.get_string('redis-channel-group', 'tasks')

    @property
    def expose_rest_api(self) -> bool:
        """Channel-tasks setting: whether to deploy the DRF API.

        Defaults to `False`."""
        return self.get_boolean('expose-rest-api', False)



"""
This module provides the tools to import and validate task coroutine functions specified by their
dotted path. See its main function :py:func:`django_tasks.task_inspector.get_task_coro`.
"""
from __future__ import annotations

import collections
import inspect
import importlib

from typing import Callable, Coroutine

from rest_framework import exceptions

from django_tasks.typing import JSON


class TaskCoroutine:
    def __init__(self, registered_task: str, **inputs: JSON):
        self.inputs = inputs
        self.errors: dict[str, list[str]] = collections.defaultdict(list)
        self.set_callable(registered_task)

    @property
    def coroutine(self) -> Coroutine:
        """The coroutine instance ready to run in event loop."""
        return self.callable(**self.inputs)

    def set_callable(self, registered_task: str) -> None:
        if '.' not in registered_task:
            self.errors['registered_task'].append(f"Missing module name in import path '{registered_task}'.")
        else:
            module_path, name = registered_task.strip().rsplit('.', 1)

            try:
                module = importlib.import_module(module_path)
            except ImportError:
                self.errors['registered_task'].append(f"Cannot import module '{module_path}'.")
            else:
                callable = getattr(module, name, None)
                if inspect.iscoroutinefunction(callable):
                    self.check_inputs(callable)
                    self.callable: Callable[..., Coroutine] = callable
                else:
                    self.errors['registered_task'].append(f"Referenced object {callable} is not a coroutine function.")

    def check_inputs(self, callable: Callable):
        params = inspect.signature(callable).parameters
        required_keys = set(k for k, v in params.items() if v.default == inspect._empty)
        optional_keys = set(k for k, v in params.items() if v.default != inspect._empty)

        input_keys = set(self.inputs)
        missing_keys = required_keys - input_keys
        unknown_keys = input_keys - required_keys - optional_keys

        if missing_keys:
            self.errors['inputs'].append(f'Missing required parameters {missing_keys}.')

        if unknown_keys:
            self.errors['inputs'].append(f'Unknown parameters {unknown_keys}.')


def get_task_coro(registered_task: str, inputs: dict[str, JSON]) -> TaskCoroutine:
    """
    Tries to obtain a registered task coroutine taking the given inputs; raises a
    :py:class:`rest_framework.exceptions.ValidationError` on failure.

    :param registered_task: The full import dotted path of the coroutine function.
    :param inputs: Input parameters for the coroutine function.
    """
    task_coro = TaskCoroutine(registered_task, **inputs)

    if task_coro.errors:
        raise exceptions.ValidationError(task_coro.errors)

    return task_coro



"""This module provides the :py:class:`django_tasks.task_runner.TaskRunner` class."""
from __future__ import annotations

import asyncio
import inspect
import logging
import time
import threading

from typing import Callable, Coroutine

from channels.layers import get_channel_layer
from rest_framework import status
from django.conf import settings

from django_tasks.task_cache import TaskCache
from django_tasks.typing import TaskStatusJSON, TaskMessageJSON


class TaskRunner:
    """
    Singleton class in charge of scheduling background task runs with `asyncio`, in a dedicated (worker) thread,
    and of broadcasting the task states and results through a configured channel layer.

    The worker thread is a daemon thread running a separate (worker) event loop; that worker loop runs
    all the scheduled background tasks concurrently.

    Usage, in async context, is as follows::

        runner = TaskRunner.get()
        future = await runner.schedule(coroutine, *callbacks, task_id='<UniversalID>', user_name='<DjangoUsername>')

    See the :py:meth:`TaskRunner.schedule` method.
    """

    #: Will hold the list of existing :py:class:`TaskRunner` instances, containing at most one instance.
    instances: list[TaskRunner] = []

    @classmethod
    def get(cls) -> TaskRunner:
        """
        Returns the last instance created, a new one if necessary, and ensures that its worker thread is alive.
        """
        if not cls.instances:
            cls()

        cls.instances[-1].ensure_alive()

        logging.getLogger('django').debug('Using task runner: %s.', cls.instances[-1])
        return cls.instances[-1]

    def __init__(self):
        caller_frame = inspect.stack()[1][0]
        caller_name = (f"{caller_frame.f_locals['cls'].__name__}.{caller_frame.f_code.co_name}"
                       if 'cls' in caller_frame.f_locals else caller_frame.f_code.co_name)
        assert caller_name == 'TaskRunner.get', (
                f"TaskRunner instances must be created by its 'get' classmethod (so not by '{caller_name}')")

        self.worker_event_loop = asyncio.new_event_loop()
        self.worker_event_loop.set_debug(settings.DEBUG)
        self.worker_thread = threading.Thread(target=self.worker_event_loop.run_forever, daemon=True)
        self.__class__.instances.append(self)
        logging.getLogger('django').debug('New task runner: %s.', self)

    def __str__(self) -> str:
        return f'loop={self.worker_event_loop}, thread={self.worker_thread}'

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}: {self}>'

    def ensure_alive(self):
        """Ensures the worker thread is alive."""
        if not self.worker_thread.is_alive():
            self.worker_thread.start()

    def run_coroutine(self, coroutine: Coroutine) -> asyncio.Future:
        """Runs the given coroutine thread-safe in the worker loop."""
        return asyncio.wrap_future(asyncio.run_coroutine_threadsafe(coroutine, self.worker_event_loop))

    def run_on_task_info(self,
                         async_callback: Callable[[str, TaskStatusJSON], Coroutine],
                         task_id: str,
                         task: asyncio.Future) -> asyncio.Future:
        """
        Runs the given async callback on the task result, taking the task ID and yielded task data as arguments.
        """
        return self.run_coroutine(async_callback(task_id, self.get_task_status(task)))

    async def schedule(self,
                       coroutine: Coroutine,
                       *coro_callbacks: Callable[[str, TaskStatusJSON], Coroutine],
                       task_id: str = '',
                       user_name: str = '') -> asyncio.Future:
        """
        Schedules the given coroutine and (optional) callbacks to run immediately in the worker thread,
        and notifies the specified user of the task state.

        :param coroutine: Coroutine of the main task to run.
        :param coro_callbacks: Sequence of async callbacks that will run just after the main task has finished,
            taking as arguments the task ID and the yielded task data.
        :param task_id: A universally unique identifier of the main task.
        :param user_name: A Django user name.

        Note that `task_id`, `user_name` are optional parameters here; this class is not responsible
        of validating them.
        """
        task_name = coroutine.__name__
        task = self.run_coroutine(coroutine)
        await self.broadcast_task(task_name, task_id, user_name, task)

        task.add_done_callback(lambda tk: self.run_coroutine(self.broadcast_task(task_name, task_id, user_name, tk)))

        for coro_callback in coro_callbacks:
            task.add_done_callback(lambda tk: self.run_on_task_info(coro_callback, task_id, tk))

        return task

    @classmethod
    async def broadcast_task(cls, name: str, task_id: str, user_name: str, task: asyncio.Future):
        """Caches the task information, and sends it to all consumers to which the user is connected,
        specifying a message type per task status.

        :param name: The name of the task coroutine.
        :param task_id: The universally unique identifier of the task.
        :param user_name: The name of the Django user that will get noticed.
        :param task: The task future.

        Note that this class is not responsible of validating these parameters.
        """
        status_data = cls.get_task_status(task)
        status_data['registered_task'] = name
        task_info: TaskMessageJSON = {'task_id': task_id, 'detail': status_data}

        user_task_cache = TaskCache(user_name)
        user_task_cache.cache_task_event(task_id, task_info)

        task_event = {'type': f"task.{status_data['status'].lower()}", 'content': task_info, 'timestamp': time.time()}
        channel_layer = get_channel_layer()
        await channel_layer.group_send(f'{user_name}_{settings.CHANNEL_TASKS.channel_group}', task_event)

    @staticmethod
    def get_task_status(task: asyncio.Future) -> TaskStatusJSON:
        """Extracts and returns the corresponding task status and result (if any result)."""
        task_info: TaskStatusJSON = {'http_status': status.HTTP_200_OK}

        if not task.done():
            task_info['status'] = 'Started'
        elif task.cancelled():
            task_info['status'] = 'Cancelled'
        elif task.exception():
            task_info.update({'status': 'Error',
                              'exception-repr': repr(task.exception())})
        else:
            task_info.update({'status': 'Success', 'output': task.result()})

        return task_info



import asyncio
import logging
import time

from django_tasks.admin_tools import ModelTask


async def sleep_test(duration, raise_error=False):
    logging.getLogger('django').info('Starting sleep test.')
    await asyncio.sleep(duration)

    if raise_error:
        logging.getLogger('django').info('Sleep test done with raise.')
        raise Exception('Test error')

    logging.getLogger('django').info('Sleep test done with no raise.')
    return f"Slept for {duration} seconds"


async def doctask_access_test(instance_ids: list[int]):
    def log_retrieved(doctask):
        time.sleep(1)
        logging.getLogger('django').info('Retrieved %s', repr(doctask))
        time.sleep(1)

    await ModelTask('django_tasks', 'DocTask', log_retrieved)(instance_ids)
    await asyncio.sleep(4)


async def doctask_deletion_test(instance_ids: list[int]):
    def delete_doctasks(doctask):
        doctask.delete()
        logging.getLogger('django').info('Deleted %s', repr(doctask))
        time.sleep(1)

    await ModelTask('django_tasks', 'DocTask', delete_doctasks)(instance_ids)



"""This module provides the DRF serializers, which are employed in web-socket and HTTP endpoints."""
from __future__ import annotations

import logging

from asgiref.sync import sync_to_async
from typing import Any

from adrf.serializers import ModelSerializer
from rest_framework.serializers import SlugRelatedField

from django_tasks import models

from django_tasks.task_inspector import get_task_coro
from django_tasks.typing import JSON, TaskJSON


class DocTaskSerializer(ModelSerializer):
    """Model serializer for the :py:class:`django_tasks.models.DocTask` model."""
    registered_task = SlugRelatedField(
        slug_field='dotted_path', queryset=models.RegisteredTask.objects.all())

    class Meta:
        model = models.DocTask
        read_only_fields = ('id', 'scheduled_at', 'completed_at', 'document')
        fields = ('registered_task', 'inputs', *read_only_fields)

    @classmethod
    async def get_valid_task_group_serializer(cls, json_content: JSON, *args, **kwargs) -> DocTaskSerializer:
        """
        Creates and returns a valid serializer instance for the given array of doc-task data.
        Raises a :py:class:`rest_framework.exceptions.ValidationError` on failure.
        """
        kwargs.update(dict(many=True, data=json_content))
        many_serializer = cls(*args, **kwargs)
        logging.getLogger('django').debug('Validating task data: json_content=%s.', json_content)
        await sync_to_async(many_serializer.is_valid)(raise_exception=True)

        return many_serializer

    @classmethod
    async def create_doctask_group(cls,
                                   json_content: JSON,
                                   *args, **kwargs) -> tuple[DocTaskSerializer, list[models.DocTask]]:
        """
        Creates an array of :py:class:`django_tasks.models.DocTask` instances for the given array of
        doc-task data.
        Raises a :py:class:`rest_framework.exceptions.ValidationError` on failure.
        """
        many_serializer = await cls.get_valid_task_group_serializer(json_content, *args, **kwargs)
        doctasks = await many_serializer.asave()

        return many_serializer, doctasks

    @classmethod
    async def get_valid_task_data(cls, json_content: JSON, *args, **kwargs) -> TaskJSON:
        kwargs['data'] = json_content
        serializer = cls(*args, **kwargs)
        logging.getLogger('django').debug('Validating task data: %s.', json_content)
        await sync_to_async(serializer.is_valid)(raise_exception=True)
        data: TaskJSON = await serializer.adata
        return data

    @classmethod
    async def get_valid_task_array_data(cls, json_content: JSON, *args, **kwargs) -> list[TaskJSON]:
        many_serializer = await cls.get_valid_task_group_serializer(json_content, *args, **kwargs)
        data: list[TaskJSON] = await many_serializer.adata
        return data

    def validate(self, attrs: dict[str, Any]) -> dict[str, Any]:
        """
        Performs the validation of the task coroutine function of the specified
        :py:class:`django_tasks.models.RegisteredTask` with the given input parameters.
        Raises a :py:class:`rest_framework.exceptions.ValidationError` on failure.
        """
        self.context['task_coro'] = get_task_coro(str(attrs['registered_task']), attrs['inputs'])

        return attrs
